// This code provides a set of functions that can be called from inside
// NVIDIA CUDA Kernels.
#include "GPUgaussMLEv2.h"

#ifndef GPUGAUSSLIB_H
#define GPUGAUSSLIB_H

__device__ float kernel_IntGauss1D(int ii, float x, float sigma) {
    float norm=1.0/2.0/sigma/sigma;
    return 1.0/2.0*(erf((ii-x+0.5)*sqrt(norm))-erf((ii-x-0.5)*sqrt(norm)));
}

__device__ float kernel_alpha(float z, float Ax, float Bx, float d){
    return 1+pow(z/d, 2)+Ax*pow(z/d, 3)+Bx*pow(z/d, 4);
}

__device__ float kernel_dalphadz(float z, float Ax, float Bx, float d){
    return (2.0*z/(d*d) + 3.0*Ax*pow(z, 2)/(d*d*d)+4.0*Bx*pow(z, 3)/pow(d, 4));
}

__device__ float kernel_d2alphadz2(float z, float Ax, float Bx, float d){
    return (2.0/(d*d) + 6.0*Ax*z/(d*d*d)+12.0*Bx*pow(z, 2)/pow(d, 4));
}

__device__ void kernel_DerivativeIntGauss1D(int ii, float x, float sigma, float N,
        float PSFy, float *dudt, float *d2udt2) {
    
    float a, b;
    a = exp(-1.0/2.0*pow(((ii+0.5-x)/sigma), 2.0));
    b = exp(-1.0/2.0*pow((ii-0.5-x)/sigma, 2.0));
    
    *dudt = -N/sqrt(2.0*pi)/sigma*(a-b)*PSFy;
    
    if (d2udt2)
        *d2udt2 =-N/sqrt(2.0*pi)/pow(sigma, 3)*((ii+0.5-x)*a-(ii-0.5-x)*b)*PSFy;
}

__device__ void kernel_DerivativeIntGauss1DSigma(int ii, float x, 
        float Sx, float N, float PSFy, float *dudt, float *d2udt2) {
    
    float ax, bx;
    
    ax = exp(-1.0/2.0*pow(((ii+0.5-x)/Sx), 2.0));
    bx = exp(-1.0/2.0*pow((ii-0.5-x)/Sx, 2.0)); 
    *dudt = -N/sqrt(2.0*pi)/Sx/Sx*(ax*(ii-x+0.5)-bx*(ii-x-0.5))*PSFy;
    
    if (d2udt2)
        *d2udt2 =-2.0/Sx*dudt[0]-N/sqrt(2.0*pi)/pow(Sx, 5)*(ax*pow((ii-x+0.5), 3)-bx*pow((ii-x-0.5), 3))*PSFy;
}

__device__ void kernel_DerivativeIntGauss2DSigma(int ii, int jj, float x, float y,
        float S, float N, float PSFx, float PSFy, float *dudt, float *d2udt2) {
    
    float dSx, dSy, ddSx, ddSy;
    
    kernel_DerivativeIntGauss1DSigma(ii,x,S,N,PSFy,&dSx,&ddSx);
    kernel_DerivativeIntGauss1DSigma(jj,y,S,N,PSFx,&dSy,&ddSy);
   
    *dudt    = dSx+dSy;
    if (d2udt2) *d2udt2 =ddSx+ddSy;
    
}

__device__ void kernel_DerivativeIntGauss2Dz(int ii, int jj, float *theta,
        float PSFSigma_x,float PSFSigma_y, float Ax, float Ay, float Bx, float By, float gamma, float d, float *pPSFx, float *pPSFy, float *dudt, float *d2udt2) {
    
    float Sx, Sy, dSx, dSy, ddSx, ddSy, dSdzx, dSdzy,ddSddzx,ddSddzy;
    float z, PSFx, PSFy,alphax,alphay,ddx,ddy;
    float dSdalpha_x,dSdalpha_y,d2Sdalpha2_x,d2Sdalpha2_y;
    z=theta[4];
    
    alphax  = kernel_alpha(z-gamma, Ax, Bx, d);
    alphay  = kernel_alpha(z+gamma, Ax, Bx, d);
 
    Sx=PSFSigma_x*sqrt(alphax);
    Sy=PSFSigma_y*sqrt(alphay);
    
    PSFx=kernel_IntGauss1D(ii, theta[0], Sx);
    PSFy=kernel_IntGauss1D(jj, theta[1], Sy);
    *pPSFx=PSFx;
    *pPSFy=PSFy;
    
    kernel_DerivativeIntGauss1D(ii, theta[0], Sx, theta[2], PSFy, &dudt[0], &ddx);
    kernel_DerivativeIntGauss1D(jj, theta[1], Sy, theta[2], PSFx, &dudt[1], &ddy);
    kernel_DerivativeIntGauss1DSigma(ii, theta[0], Sx, theta[2], PSFy, &dSx, &ddSx);
    kernel_DerivativeIntGauss1DSigma(jj, theta[1], Sy, theta[2], PSFx, &dSy, &ddSy);

    dSdalpha_x=PSFSigma_x/2.0/sqrt(alphax);
    dSdalpha_y=PSFSigma_y/2.0/sqrt(alphay);
    
    dSdzx  = dSdalpha_x*kernel_dalphadz(z-gamma, Ax, Bx, d); 
    dSdzy  = dSdalpha_y*kernel_dalphadz(z+gamma, Ay, By, d);
    dudt[4] = dSx*dSdzx+dSy*dSdzy;
    
    if (d2udt2){
    d2udt2[0] =ddx;
    d2udt2[1] =ddy;
    
    d2Sdalpha2_x=-PSFSigma_x/4.0/pow(alphax,1.5f);
    d2Sdalpha2_y=-PSFSigma_y/4.0/pow(alphay,1.5f);
    
    ddSddzx  = d2Sdalpha2_x*pow(kernel_dalphadz(z-gamma, Ax, Bx, d),2)+dSdalpha_x*kernel_d2alphadz2(z-gamma, Ax, Bx, d); 
    ddSddzy  = d2Sdalpha2_y*pow(kernel_dalphadz(z+gamma, Ax, Bx, d),2)+dSdalpha_y*kernel_d2alphadz2(z+gamma, Ay, By, d); 
    
    d2udt2[4] =ddSx*(dSdzx*dSdzx)+dSx*ddSddzx+
            ddSy*(dSdzy*dSdzy)+dSy*ddSddzy;
    }
}

__device__ void kernel_CenterofMass2D(int sz, float * data, float *x, float *y) {
    
    float tmpx=0.0;
    float tmpy=0.0;
    float tmpsum=0.0;
    int ii, jj;
    
    for (ii=0;ii<sz;ii++) for(jj=0;jj<sz;jj++) {
        tmpx+=data[sz*jj+ii]*ii;
        tmpy+=data[sz*jj+ii]*jj;
        tmpsum+=data[sz*jj+ii];
    }
    *x=tmpx/tmpsum;
    *y=tmpy/tmpsum;
}

__device__ void kernel_GaussFMaxMin2D(int sz, float sigma, float * data, float *MaxN, float *MinBG) {
    
    int ii, jj, kk, ll;
    float filteredpixel, sum;
    *MaxN=0.0;
    *MinBG=10e10; //big
    
    float norm=1.0/2.0/sigma/sigma;
    //loop over all pixels
    for (kk=0;kk<sz;kk++) for (ll=0;ll<sz;ll++){
        filteredpixel=0.0;
        sum=0.0;
        for (ii=0;ii<sz;ii++) for(jj=0;jj<sz;jj++){
            filteredpixel+=exp(-pow((float)(ii-kk), 2)*norm)*exp(-pow((float)(ll-jj), 2)*norm)*data[ii*sz+jj];
            sum+=exp(-pow((float)(ii-kk), 2)*norm)*exp(-pow((float)(ll-jj), 2)*norm);
        }
        filteredpixel/=sum;
        
        *MaxN=max(*MaxN, filteredpixel);
        *MinBG=min(*MinBG, filteredpixel);
    }
}

#endif
















