/* C++ wrapper for GPUgaussMLEv2.cu GPU-kernel
 * input:
 *   float *data = input data, pointer to matrix
 *   int sz = size of array; x=y (array must be square)
 *   int Nfitraw = number of arrays
 *   int PSFSigma = point spread function sigma
 *   int iterations = number of iterations
 *   int fittype
 *   [if fittype==3 :
 *     float Ax, Ay, Bx, By, gamma, d, PSFSigma_y = parameters needed for fittype 3]
 * output:
 *   float *d_Parameters = output parameters, size *Nfitraw*<number of parameters> (=NV_P|NV_PS|NV_PS2)
 *   float *d_CRLBs = output CRLBs, size *Nfitraw*<number of parameters> (=NV_P|NV_PS|NV_PS2)
 *   float *d_LogLikelihood = output loglikelihood, size *Nfitraw
 */

#include <stdio.h>
#include <stdarg.h> // for var_arg
#include <stdlib.h>
#include <stdexcept>

#include <math.h>  // for floor
#include <Windows.h> // for min() and max()

#include "cuda_runtime.h"
#include "cpp_GPUgaussMLEv2.h"

#include "..\GPUgaussMLEv2\GPUgaussMLEv2.cu"

using namespace std;

void CUDAERRROR(const char *instr) {
    cudaError_t errornum;
    const char *str;
    if (errornum=cudaGetLastError()) {
        str=cudaGetErrorString(errornum);
        printf("%s\n", str);
    }
}

extern "C" __declspec(dllexport)
void gpugaussmle(float *data, int sz, int Nfitraw, float PSFSigma, int iterations, int fittype,
								float *Parameters, float *CRLBs, float *LogLikelihood, 
								...) {
    int blockx;
    int threadx;
	float *d_data;
	int Nfits;
	float *d_Parameters, *d_CRLBs, *d_LogLikelihood;
    float Ax, Ay, Bx, By, gamma, d, PSFSigma_y;

	printf("sz: %d, Nfitraw: %d, PSFSigma: %f, iterations: %d, fittype: %d\n", sz, Nfitraw, PSFSigma, iterations, fittype);

	//sanity check
	if (sz > IMSZBIG) {
		printf("X,Y dimensions of data must be smaller than %d.\n", IMSZBIG);
		exit(1);
	}

    //put as many images as fit into a block
    int BlockSize = floor((float)15000/4/sz/sz);
    BlockSize = max(4, BlockSize);
    BlockSize = min(BSZ, BlockSize);

    Nfits=BlockSize*ceil( (float) Nfitraw/BlockSize);
    Nfits=Nfitraw;  ///WTF?

	if (fittype==3) {
		va_list args;
		va_start(args, *LogLikelihood);
        Ax = float va_arg(args, double); //http://c-faq.com/varargs/float.html float's not allowed in va_arg
        Ay = float va_arg(args, double);
        Bx = float va_arg(args, double);
        By = float va_arg(args, double);
        gamma = float va_arg(args, double);
        d = float va_arg(args, double);
        PSFSigma_y = float va_arg(args, double);
		va_end(args);
		printf("Ax: %f, Ay: %f, Bx: %f, By: %f, gamma: %f, d: %f, PSFSigma_y: %f\n", Ax, Ay, Bx, By, gamma, d, PSFSigma_y);
	}
	
	//create device variable for data and copy to device
    cudaMalloc((void**)&d_data, sz*sz*Nfits*sizeof(float));
    cudaMemset(d_data, 0, sz*sz*Nfits*sizeof(float));
    cudaMemcpy(d_data, data, sz*sz*Nfitraw*sizeof(float), cudaMemcpyHostToDevice);
    
	//create output for parameters and CRLBs
    switch(fittype){
        case 1: // (x,y,bg,I)
            cudaMalloc((void**)&d_Parameters,   NV_P*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_P*Nfits*sizeof(float));
            break;
        case 2: // (x,y,bg,I,Sigma)
        case 3: // (x,y,bg,I,z)
            cudaMalloc((void**)&d_Parameters,   NV_PS*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_PS*Nfits*sizeof(float));
            break;      
        case 4: // (x,y,bg,I,Sx,Sy)
            cudaMalloc((void**)&d_Parameters,   NV_PS2*Nfits*sizeof(float));
            cudaMalloc((void**)&d_CRLBs,        NV_PS2*Nfits*sizeof(float));
            break;
    }
    cudaMalloc((void**)&d_LogLikelihood,        Nfits*sizeof(float));

    //setup kernel
    blockx = ceil( (float)Nfits/(float)BlockSize);
    threadx= BlockSize;
    
    dim3 dimBlock(threadx);
    dim3 dimGrid(blockx);
    
    printf("threadx: %d, blockx: %d, Nfitraw: %d\n", threadx, blockx, Nfitraw);
    
    switch(fittype) {
        case 1: //fit x,y,bg,I
            kernel_MLEFit<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            break;
            
        case 2: //fit x,y,bg,I,sigma
            kernel_MLEFit_sigma<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            break;
            
        case 3: //fit x,y,bg,I,z
            kernel_MLEFit_z<<<dimGrid, dimBlock>>>(d_data, PSFSigma, Ax, Ay, Bx, By, gamma, d, PSFSigma_y, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            break;
            
        case 4: //fit x,y,bg,I,sigmax,sigmay
            kernel_MLEFit_sigmaxy<<<dimGrid, dimBlock>>>(d_data, PSFSigma, sz, iterations, d_Parameters, d_CRLBs, d_LogLikelihood, Nfits);
            
    }
    CUDAERRROR("kernel");

    //copy to output
    switch(fittype){
        case 1: // (x,y,bg,I)
            cudaMemcpy((float *)Parameters, d_Parameters, NV_P*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)CRLBs, d_CRLBs, NV_P*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
        case 2: // (x,y,bg,I,Sigma)
        case 3: // (x,y,bg,I,z)
            cudaMemcpy((float *)Parameters, d_Parameters, NV_PS*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)CRLBs, d_CRLBs, NV_PS*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
        case 4: // (x,y,bg,I,Sx,Sy)
            cudaMemcpy((float *)Parameters, d_Parameters, NV_PS2*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            cudaMemcpy((float *)CRLBs, d_CRLBs, NV_PS2*Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);
            break;
    }
    cudaMemcpy((float *)LogLikelihood, d_LogLikelihood, Nfitraw*sizeof(float), cudaMemcpyDeviceToHost);

	//cleanup
    cudaFree(d_Parameters);
    cudaFree(d_CRLBs);
    cudaFree(d_LogLikelihood);
    
}
